# THIS FILE IS AUTO-GENERATED. PLEASE DO NOT MODIFY# version file for ldap3
# generated on 2016-11-18 08:08:03.830784
# on system uname_result(system='Windows', node='ELITE10GC', release='10', version='10.0.14393', machine='AMD64', processor='Intel64 Family 6 Model 58 Stepping 9, GenuineIntel')
# with Python 3.5.2 - ('v3.5.2:4def2a2901a5', 'Jun 25 2016 22:18:55') - MSC v.1900 64 bit (AMD64)
#
__version__ = '2.1.1'
__author__ = 'Giovanni Cannata'
__email__ = 'cannatag@gmail.com'
__url__ = 'https://github.com/cannatag/ldap3'
__description__ = 'A strictly RFC 4510 conforming LDAP V3 pure Python client library. The same codebase works with Python 2, Python 3, PyPy and PyPy3'
__status__ = '5 - Production/Stable'
__license__ = 'LGPL v3'
